package love.distributedrebirth.no2all.nostr.fire;

import java.io.StringReader;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Consumer;

import jakarta.json.Json;
import jakarta.json.JsonArray;
import jakarta.json.JsonReader;
import jakarta.json.JsonValue;
import love.distributedrebirth.no2all.nostr.model.message.NoStrMsgToClient;
import love.distributedrebirth.no2all.nostr.model.message.NoStrMsgToRelay;
import love.distributedrebirth.no2all.nostr.nip.NoStrImplMessageType;

public final class NoStrFirePipe {
	private static final String ERROR_INVALID_JSON = "Invalid json; "; // on json parse
	private static final String ERROR_INVALID_ROOT = "Invalid root; "; // on none array root
	private static final String ERROR_TYPE_UNKNOWN = "Unknown type; "; // on not in values()
	private static final String ERROR_TYPE_ILLEGAL = "Illegal type"; // on null mapper
	private static final int MSG_TYPE_INDEX = 0;
	private final Consumer<String> errHandler;
	private final Consumer<NoStrMsgToRelay> msgHandlerToRelay;
	private final Consumer<NoStrMsgToClient> msgHandlerToClient;

	public NoStrFirePipe(Consumer<String> errHandler, Consumer<NoStrMsgToRelay> msgHandlerToRelay,
			Consumer<NoStrMsgToClient> msgHandlerToClient) {
		this.errHandler = Objects.requireNonNull(errHandler);
		this.msgHandlerToRelay = msgHandlerToRelay;
		this.msgHandlerToClient = msgHandlerToClient;
		if (msgHandlerToRelay != null && msgHandlerToClient != null) {
			throw new IllegalArgumentException("Only use one message handler");
		}
		if (msgHandlerToRelay == null && msgHandlerToClient == null) {
			throw new NullPointerException("Need one message handler");
		}
	}

	public void onFire(String msgStr) {
		JsonValue msgJson = null;
		try (JsonReader jsonReader = Json.createReader(new StringReader(msgStr))) {
			msgJson = jsonReader.readValue();
		} catch (Exception e) {
			errHandler.accept(ERROR_INVALID_JSON + e.getMessage());
			return;
		}
		if (!JsonValue.ValueType.ARRAY.equals(msgJson.getValueType())) {
			errHandler.accept(ERROR_INVALID_ROOT + msgJson.getValueType().name());
			return;
		}
		JsonArray msgPacket = msgJson.asJsonArray();
		String msgTypeRaw = msgPacket.getString(MSG_TYPE_INDEX);
		if (!NoStrImplMessageType.valueOfValid(msgTypeRaw)) {
			errHandler.accept(ERROR_TYPE_UNKNOWN + msgTypeRaw);
			return;
		}
		NoStrImplMessageType msgType = NoStrImplMessageType.valueOf(msgTypeRaw);
		if (msgHandlerToClient != null) {
			acceptType(msgType.getMapperToClient(), v -> msgHandlerToClient.accept(v.apply(msgPacket))); // on relay
		} else {
			acceptType(msgType.getMapperToRelay(), v -> msgHandlerToRelay.accept(v.apply(msgPacket))); // on client
		}
	}

	private <T> void acceptType(Optional<T> mapper, Consumer<T> messageHandler) {
		if (mapper.isEmpty()) {
			errHandler.accept(ERROR_TYPE_ILLEGAL);
			return;
		}
		messageHandler.accept(mapper.get());
	}
}
