package love.distributedrebirth.no2all.octo.trust;

import java.math.BigInteger;
import java.util.Arrays;

public final class OctoTrustSchnorr {

	public static byte[] sign(byte[] msg, byte[] secKey, byte[] auxRand) {
		if (msg.length != 32) {
			throw new IllegalArgumentException("The message must be a 32-byte array.");
		}
		BigInteger secKey0 = bytesToBigInt(secKey);
		if (!(BigInteger.ONE.compareTo(secKey0) <= 0
				&& secKey0.compareTo(OctoTrustSchnorrPoint.getn().subtract(BigInteger.ONE)) <= 0)) {
			throw new IllegalArgumentException("The secret key must be an integer in the range 1..n-1.");
		}
		OctoTrustSchnorrPoint P = OctoTrustSchnorrPoint.mul(OctoTrustSchnorrPoint.getG(), secKey0);
		if (!P.hasEvenY()) {
			secKey0 = OctoTrustSchnorrPoint.getn().subtract(secKey0);
		}
		int len = bigIntToBytes(secKey0).length + P.toBytes().length + msg.length;
		byte[] buf = new byte[len];
		byte[] t = xor(bigIntToBytes(secKey0), OctoTrustSchnorrPoint.taggedHash("BIP0340/aux", auxRand));
		System.arraycopy(t, 0, buf, 0, t.length);
		System.arraycopy(P.toBytes(), 0, buf, t.length, P.toBytes().length);
		System.arraycopy(msg, 0, buf, t.length + P.toBytes().length, msg.length);
		BigInteger k0 = bytesToBigInt(OctoTrustSchnorrPoint.taggedHash("BIP0340/nonce", buf))
				.mod(OctoTrustSchnorrPoint.getn());
		if (k0.compareTo(BigInteger.ZERO) == 0) {
			throw new IllegalArgumentException("Failure. This happens only with negligible probability.");
		}
		OctoTrustSchnorrPoint R = OctoTrustSchnorrPoint.mul(OctoTrustSchnorrPoint.getG(), k0);
		BigInteger k;
		if (!R.hasEvenY()) {
			k = OctoTrustSchnorrPoint.getn().subtract(k0);
		} else {
			k = k0;
		}
		len = R.toBytes().length + P.toBytes().length + msg.length;
		buf = new byte[len];
		System.arraycopy(R.toBytes(), 0, buf, 0, R.toBytes().length);
		System.arraycopy(P.toBytes(), 0, buf, R.toBytes().length, P.toBytes().length);
		System.arraycopy(msg, 0, buf, R.toBytes().length + P.toBytes().length, msg.length);
		BigInteger e = bytesToBigInt(OctoTrustSchnorrPoint.taggedHash("BIP0340/challenge", buf))
				.mod(OctoTrustSchnorrPoint.getn());
		BigInteger kes = k.add(e.multiply(secKey0)).mod(OctoTrustSchnorrPoint.getn());
		len = R.toBytes().length + bigIntToBytes(kes).length;
		byte[] sig = new byte[len];
		System.arraycopy(R.toBytes(), 0, sig, 0, R.toBytes().length);
		System.arraycopy(bigIntToBytes(kes), 0, sig, R.toBytes().length, bigIntToBytes(kes).length);
		if (!verify(msg, P.toBytes(), sig)) {
			throw new IllegalStateException("The signature does not pass verification.");
		}
		return sig;
	}

	public static boolean verify(byte[] msg, byte[] pubkey, byte[] sig) {
		if (msg.length != 32) {
			throw new IllegalArgumentException("The message must be a 32-byte array.");
		}
		if (pubkey.length != 32) {
			throw new IllegalArgumentException("The public key must be a 32-byte array.");
		}
		if (sig.length != 64) {
			throw new IllegalArgumentException("The signature must be a 64-byte array.");
		}
		OctoTrustSchnorrPoint P = OctoTrustSchnorrPoint.liftX(pubkey);
		if (P == null) {
			return false;
		}
		BigInteger r = bytesToBigInt(Arrays.copyOfRange(sig, 0, 32));
		BigInteger s = bytesToBigInt(Arrays.copyOfRange(sig, 32, 64));
		if (r.compareTo(OctoTrustSchnorrPoint.getp()) >= 0 || s.compareTo(OctoTrustSchnorrPoint.getn()) >= 0) {
			return false;
		}
		int len = 32 + pubkey.length + msg.length;
		byte[] buf = new byte[len];
		System.arraycopy(sig, 0, buf, 0, 32);
		System.arraycopy(pubkey, 0, buf, 32, pubkey.length);
		System.arraycopy(msg, 0, buf, 32 + pubkey.length, msg.length);
		BigInteger e = bytesToBigInt(OctoTrustSchnorrPoint.taggedHash("BIP0340/challenge", buf))
				.mod(OctoTrustSchnorrPoint.getn());
		OctoTrustSchnorrPoint R = OctoTrustSchnorrPoint.add(OctoTrustSchnorrPoint.mul(OctoTrustSchnorrPoint.getG(), s),
				OctoTrustSchnorrPoint.mul(P, OctoTrustSchnorrPoint.getn().subtract(e)));
		return R != null && R.hasEvenY() && R.getX().compareTo(r) == 0;
	}

	public static byte[] extractPublicKey(byte[] secKey) {
		BigInteger x = bytesToBigInt(secKey);
		if (!(BigInteger.ONE.compareTo(x) <= 0
				&& x.compareTo(OctoTrustSchnorrPoint.getn().subtract(BigInteger.ONE)) <= 0)) {
			throw new IllegalArgumentException("The secret key must be an integer in the range 1..n-1.");
		}
		OctoTrustSchnorrPoint result = OctoTrustSchnorrPoint.mul(OctoTrustSchnorrPoint.G, x);
		return OctoTrustSchnorrPoint.bytesFromPoint(result);
	}

	protected static byte[] xor(byte[] b0, byte[] b1) {
		if (b0.length != b1.length) {
			return null;
		}
		byte[] result = new byte[b0.length];
		int i = 0;
		for (byte b : b0) {
			result[i] = (byte) (b ^ b1[i]);
			i++;
		}
		return result;
	}

	protected static byte[] bigIntToBytes(BigInteger n) {
		byte[] b = n.toByteArray();
		if (b.length == 32) {
			return b;
		} else if (b.length > 32) {
			return Arrays.copyOfRange(b, b.length - 32, b.length);
		} else {
			byte[] buf = new byte[32];
			System.arraycopy(b, 0, buf, buf.length - b.length, b.length);
			return buf;
		}
	}

	protected static BigInteger bytesToBigInt(byte[] b) {
		return new BigInteger(1, b);
	}
}
