package love.distributedrebirth.no2all.octo.trust;

import java.math.BigInteger;

public class OctoTrustSchnorrPoint {

	final static private BigInteger p = new BigInteger(
			"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F", 16);
	final static private BigInteger n = new BigInteger(
			"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141", 16);

	final static public OctoTrustSchnorrPoint G = new OctoTrustSchnorrPoint(
			new BigInteger("79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798", 16),
			new BigInteger("483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8", 16));

	private static final BigInteger BI_TWO = BigInteger.valueOf(2);
	private final BigInteger x;
	private final BigInteger y;

	public OctoTrustSchnorrPoint(BigInteger x, BigInteger y) {
		this.x = x;
		this.y = y;
	}

	public static BigInteger getp() {
		return p;
	}

	public static BigInteger getn() {
		return n;
	}

	public static OctoTrustSchnorrPoint getG() {
		return G;
	}

	public BigInteger getX() {
		return x;
	}

	public BigInteger getY() {
		return y;
	}

	public static BigInteger getX(OctoTrustSchnorrPoint point) {
		return point.getX();
	}

	public static BigInteger getY(OctoTrustSchnorrPoint point) {
		return point.getY();
	}

	public boolean isInfinite() {
		return x == null || y == null;
	}

	public static boolean isInfinite(OctoTrustSchnorrPoint point) {
		return point.isInfinite();
	}

	public OctoTrustSchnorrPoint add(OctoTrustSchnorrPoint point) {
		return add(this, point);
	}

	public static OctoTrustSchnorrPoint add(OctoTrustSchnorrPoint point1, OctoTrustSchnorrPoint point2) {
		if ((point1 != null && point2 != null && point1.isInfinite() && point2.isInfinite())) {
			return infinityPoint();
		}
		if (point1 == null || point1.isInfinite()) {
			return point2;
		}
		if (point2 == null || point2.isInfinite()) {
			return point1;
		}
		if (point1.getX().equals(point2.getX()) && !point1.getY().equals(point2.getY())) {
			return infinityPoint();
		}
		BigInteger lam;
		if (point1.equals(point2)) {
			BigInteger base = point2.getY().multiply(BI_TWO);
			lam = (BigInteger.valueOf(3L).multiply(point1.getX()).multiply(point1.getX())
					.multiply(base.modPow(p.subtract(BI_TWO), p))).mod(p);
		} else {
			BigInteger base = point2.getX().subtract(point1.getX());
			lam = ((point2.getY().subtract(point1.getY())).multiply(base.modPow(p.subtract(BI_TWO), p))).mod(p);
		}
		BigInteger x3 = (lam.multiply(lam).subtract(point1.getX()).subtract(point2.getX())).mod(p);
		return new OctoTrustSchnorrPoint(x3, lam.multiply(point1.getX().subtract(x3)).subtract(point1.getY()).mod(p));
	}

	public OctoTrustSchnorrPoint mul(BigInteger n) {
		return mul(this, n);
	}

	public static OctoTrustSchnorrPoint mul(OctoTrustSchnorrPoint point, BigInteger n) {
		OctoTrustSchnorrPoint R = null;
		for (int i = 0; i < 256; i++) {
			if (n.shiftRight(i).and(BigInteger.ONE).compareTo(BigInteger.ZERO) > 0) {
				R = add(R, point);
			}
			point = add(point, point);
		}
		return R;
	}

	public boolean hasEvenY() {
		return hasEvenY(this);
	}

	public static boolean hasEvenY(OctoTrustSchnorrPoint point) {
		if (point == null) {
			return false;
		}
		if (point.getY() == null) { // added for junit
			return false;
		}
		return point.getY().mod(BI_TWO).compareTo(BigInteger.ZERO) == 0;
	}

	public static boolean isSquare(BigInteger x) {
		return x.modPow(p.subtract(BigInteger.ONE).mod(BI_TWO), p).longValue() == 1L;
	}

	public boolean hasSquareY() {
		return hasSquareY(this);
	}

	public static boolean hasSquareY(OctoTrustSchnorrPoint point) {
		return isSquare(point.getY());
	}

	public static byte[] taggedHash(String tag, byte[] msg) {
		byte[] tagHash = OctoTrustHash.sha256(tag.getBytes());
		int len = (tagHash.length * 2) + msg.length;
		byte[] buf = new byte[len];
		System.arraycopy(tagHash, 0, buf, 0, tagHash.length);
		System.arraycopy(tagHash, 0, buf, tagHash.length, tagHash.length);
		System.arraycopy(msg, 0, buf, tagHash.length * 2, msg.length);
		return OctoTrustHash.sha256(buf);
	}

	public static byte[] genPubKey(byte[] secKey) {
		BigInteger x = OctoTrustSchnorr.bytesToBigInt(secKey);
		if (!(BigInteger.ONE.compareTo(x) <= 0 && x.compareTo(getn().subtract(BigInteger.ONE)) <= 0)) {
			throw new IllegalArgumentException("The secret key must be an integer in the range 1..n-1.");
		}
		OctoTrustSchnorrPoint ret = OctoTrustSchnorrPoint.mul(G, x);
		return bytesFromPoint(ret);
	}

	public byte[] toBytes() {
		return bytesFromPoint(this);
	}

	public static byte[] bytesFromPoint(OctoTrustSchnorrPoint point) {
		return OctoTrustSchnorr.bigIntToBytes(point.getX());
	}

	public static OctoTrustSchnorrPoint liftX(byte[] b) {
		BigInteger x = OctoTrustSchnorr.bytesToBigInt(b);
		if (x.compareTo(p) >= 0) {
			return null;
		}
		BigInteger y_sq = x.modPow(BigInteger.valueOf(3L), p).add(BigInteger.valueOf(7L)).mod(p);
		BigInteger y = y_sq.modPow(p.add(BigInteger.ONE).divide(BigInteger.valueOf(4L)), p);
		if (y.modPow(BI_TWO, p).compareTo(y_sq) != 0) {
			return null;
		} else {
			return new OctoTrustSchnorrPoint(x, y.and(BigInteger.ONE).compareTo(BigInteger.ZERO) == 0 ? y : p.subtract(y));
		}
	}

	public static OctoTrustSchnorrPoint infinityPoint() {
		return new OctoTrustSchnorrPoint(null, null);
	}

	public boolean equals(OctoTrustSchnorrPoint point) {
		return getX().equals(point.getX()) && getY().equals(point.getY());
	}

}
