package love.distributedrebirth.no2all.react.warp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import love.distributedrebirth.no2all.react.No2AllReactSlotLoad;
import love.distributedrebirth.no2all.react.No2AllReact;
import love.distributedrebirth.no2all.react.No2AllReactListener;
import love.distributedrebirth.no2all.react.No2AllReactSlot;
import love.distributedrebirth.no2all.react.No2AllReactSlotSignal;
import love.distributedrebirth.no2all.react.No2AllReactTypeScript;

public class No2AllReactWarpCore implements No2AllReact {

	public static final No2AllReactSlot DILITHIUM = No2AllReactSlot.ofClass(No2AllReactWarpCore.class);
	private final Map<No2AllReactSlot, No2AllReactWarpFuel> slots = new HashMap<>();
	private final Map<Class<?>, Object> services = new HashMap<>();
	private final List<BacklogEvent> backlog = new ArrayList<>();
	private Object currentEvent = null;

	public No2AllReactWarpCore() {
		claim(DILITHIUM);
		toWarpFuel(DILITHIUM).registrateTypeIn(No2AllReactSlotLoad.class);
		toWarpFuel(DILITHIUM).registrateTypeOut(No2AllArtWarpSlotAdd.class);
		toWarpFuel(DILITHIUM).registrateTypeOut(No2AllArtWarpSlotRemove.class);
	}

	public void load(No2AllReactTypeScript script) {
		Objects.requireNonNull(script).onEvent(new No2AllReactSlotSignal<>(DILITHIUM, new No2AllReactSlotLoad(), this));
	}

	public No2AllReactSlotContract getSlotContract(No2AllReactSlot slot) {
		return toWarpFuel(slot).readContract();
	}

	public List<No2AllReactSlot> listChilds(No2AllReactSlot slot) {
		Objects.requireNonNull(slot);
		List<No2AllReactSlot> result = new ArrayList<>();
		synchronized (slots) {
			for (No2AllReactSlot slug : slots.keySet()) {
				if (slot == null) {
					result.add(slug);
					continue;
				}
				if (slug.getParent() == null) {
					continue;
				}
				if (slug.getParent().getSlotPath().equals(slot.getSlotPath())) {
					result.add(slug);
				}
			}
		}
		return result;
	}

	@Override
	public void fireForEach(No2AllReactSlot slot, Object event) {
		for (No2AllReactSlot target : listChilds(slot)) {
			fire(target, event);
		}
	}

	@Override
	public void fire(No2AllReactSlot slot, Object event) {
		Objects.requireNonNull(slot);
		Objects.requireNonNull(event);
		if (currentEvent != null) {
			backlog.add(new BacklogEvent(slot, event));
			return;
		}
		currentEvent = event;
		try {
			toWarpFuel(slot).fire(event, slot);
			
			while (!backlog.isEmpty()) {
				List<BacklogEvent> backlog2 = new ArrayList<>(backlog);
				backlog.clear();
				for (BacklogEvent next : backlog2) {
					toWarpFuel(next.slot).fire(next.event, next.slot);
				}
			}
		} finally {
			currentEvent = null;
		}
	}

	static class BacklogEvent {
		No2AllReactSlot slot;
		Object event;
		public BacklogEvent(No2AllReactSlot slot, Object event) {
			this.slot = slot;
			this.event = event;
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> void registrate(No2AllReactSlot slot, Class<T> eventType, No2AllReactListener<T> listener) {
		No2AllReactListener<Object> listenerObj = (No2AllReactListener<Object>) listener; 
		toWarpFuel(slot).registrateListener(eventType, listenerObj);
	}

	@Override
	public void release(No2AllReactSlot slot) {
		toWarpFuel(slot).removeAll();
		synchronized (slots) {
			slots.remove(slot);
		}
		fire(DILITHIUM, new No2AllArtWarpSlotRemove(slot));
	}

	@Override
	public void claim(No2AllReactSlot slot) {
		Objects.requireNonNull(slot);
		synchronized (slots) {
			if (slots.keySet().contains(slot)) {
				throw new IllegalStateException("Slug already claimed: " + slot.getSlotPath());
			}
			slots.put(slot, new No2AllReactWarpFuel(this, slot));
		}
		fire(DILITHIUM, new No2AllArtWarpSlotAdd(slot));
	}

	@Override
	public void claimOut(No2AllReactSlot slot, Class<?> eventType) {
		toWarpFuel(slot).registrateTypeOut(eventType);
	}


	@Override
	public void claimIn(No2AllReactSlot slot, Class<?> eventType) {
		toWarpFuel(slot).registrateTypeIn(eventType);
	}

	@Override
	public void requireSlot(No2AllReactSlot slot, No2AllReactSlot dep) {
		if (!isClaimed(dep)) {
			throw new IllegalStateException("Script dependency missing: " + dep);
		}
		toWarpFuel(slot).addRequireSlot(dep);
	}

	@Override
	public void requireService(No2AllReactSlot slot, Class<?> serviceType) {
		toWarpFuel(slot).addRequireService(serviceType);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> T service(Class<T> serviceType) {
		Objects.requireNonNull(serviceType);
		synchronized (services) {
			return (T) services.get(serviceType);
		}
	}

	public <T> void addService(Class<T> serviceType, T service) {
		Objects.requireNonNull(serviceType);
		Objects.requireNonNull(service);
		synchronized (services) {
			services.put(serviceType, service);
		}
	}
	
	public boolean isClaimed(No2AllReactSlot slug) {
		Objects.requireNonNull(slug);
		synchronized (slots) {
			return slots.keySet().contains(slug);
		}
	}

	protected No2AllReactWarpFuel toWarpFuel(No2AllReactSlot slot) {
		Objects.requireNonNull(slot);
		No2AllReactWarpFuel result;
		synchronized (slots) {
			result = slots.get(slot);
			if (result == null) {
				throw new IllegalStateException("Slot not claimed: " + slot.getSlotPath());
			}
		}
		return result;
	}
}
