package love.distributedrebirth.no2all.wire.fetch;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.Map.Entry;

public final class WireFetchFactoryDefault implements WireFetchFactory {

	@Override
	public WireFetch buildFetcher(WireFetchResource resource, WireFetchHandler handler) {
		return new HttpDownloader(resource, handler);
	}

	static protected final class HttpDownloader implements WireFetch {

		private final WireFetchResource resource;
		private final WireFetchHandler handler;

		protected HttpDownloader(WireFetchResource resource, WireFetchHandler handler) {
			this.resource = resource;
			this.handler = handler;
		}

		@Override
		public void run() {
			try {
				handler.onStart();
				try (InputStream conn = openInputStream(resource.getUrId().toURL(), resource.getHeaders(), resource.getTimeoutMs(), 5)) {
					try (ByteArrayOutputStream result = new ByteArrayOutputStream()) {
						int totalBytes = 0;
						byte[] buffer = new byte[4096];
						int length;
						while ((length = conn.read(buffer)) != -1) {
							result.write(buffer, 0, length);
							totalBytes += length;
							handler.onProgress(totalBytes);
						}
						handler.onReady(ByteBuffer.wrap(result.toByteArray()));
					}
				}
			} catch (Exception error) {
				handler.onError(error);
			}
		}

		private InputStream openInputStream(URL url, Map<String, String> args, int timeoutMs, int redirectLimit) throws IOException {
			HttpURLConnection con = (HttpURLConnection) url.openConnection();
			con.setConnectTimeout(timeoutMs);
			con.setReadTimeout(timeoutMs);
			if (args != null) {
				for (Entry<String, String> e : args.entrySet()) {
					con.setRequestProperty(e.getKey(), e.getValue());
				}
			}
			con.connect();
			int responseCode = con.getResponseCode();
			if (responseCode < 400 && responseCode > 299) {
				// for redirects but breaks switch protocols
				redirectLimit--;
				if (redirectLimit == 0) {
					throw new IllegalStateException("Max redirect limit reached");
				}
				String redirectLocation = con.getHeaderField("Location");
				if (redirectLocation == null) {
					throw new IllegalStateException("No redirect location header");
				}
				URL redirectUrl = null;
				try {
					redirectUrl = new URL(redirectLocation);
				} catch (MalformedURLException e) {
					redirectUrl = new URL(url.getProtocol() + "://" + url.getHost() + redirectLocation);
				}
				return openInputStream(redirectUrl, args, timeoutMs, redirectLimit);
			}
			return con.getInputStream();
		}
	}
}
